@@ -7,15 +7,17 @@
 import argparse
 import os
 import re
+import time
 import types
 import shutil
 import socket
-from html2image import Html2Image
 
 from xml.etree import ElementTree
+from xml.etree.ElementTree import ParseError
 from datetime import datetime
 
 from PIL import Image, ImageDraw
+from cairosvg.surface import PNGSurface
 
 from yt_dlp import YoutubeDL
 from yt_dlp.compat import (
@@ -35,6 +37,7 @@ from yt_dlp.extractor.common import InfoExtractor
 from yt_dlp.postprocessor.ffmpeg import FFmpegPostProcessorError
 
 from bbb_dl.ffmpeg import FFMPEG
+from bbb_dl.html2image import Html2Image
 from bbb_dl.version import __version__
 
 _s = lambda p: xpath_with_ns(p, {'svg': 'http://www.w3.org/2000/svg'})
@@ -95,13 +98,33 @@ class BBBDL(InfoExtractor):
                      (?P<id>[0-9a-f\-]+)
                    '''
 
-    def __init__(self, verbose: bool, no_check_certificate: bool, encoder: str, audiocodec: str):
+    def __init__(
+        self,
+        verbose: bool,
+        no_check_certificate: bool,
+        encoder: str,
+        audiocodec: str,
+        add_webcam: bool,
+        add_annotations: bool,
+        add_cursor: bool,
+        keep_tmp_files: bool,
+        verbose_chrome: bool,
+        chrome_executable: str,
+    ):
+        self.add_webcam = add_webcam
+        self.add_annotations = add_annotations
+        self.add_cursor = add_cursor
+        self.keep_tmp_files = keep_tmp_files
+        self.verbose_chrome = verbose_chrome
+        self.chrome_executable = chrome_executable
+
         if '_VALID_URL_RE' not in self.__dict__:
             BBBDL._VALID_URL_RE = re.compile(self._VALID_URL)
 
+        self.global_retries = 10
         ydl_options = {
-            'retries': 10,
-            'fragment_retries': 10,
+            'retries': self.global_retries,
+            'fragment_retries': self.global_retries,
         }
         if verbose:
             ydl_options.update({"verbose": True})
@@ -123,9 +146,7 @@ class BBBDL(InfoExtractor):
     def _mark_watched(self, *args, **kwargs):
         return
 
-    def run(
-        self, dl_url: str, add_webcam: bool, add_annotations: bool, add_cursor, keep_tmp_files: bool, filename: str
-    ):
+    def run(self, dl_url: str, filename: str, backup: bool):
         m_obj = re.match(self._VALID_URL, dl_url)
 
         video_id = m_obj.group('id')
@@ -133,18 +154,22 @@ class BBBDL(InfoExtractor):
 
         self.to_screen("Downloading meta informations")
         # Make sure the lesson exists
-        self._download_webpage(dl_url, video_id)
+        if not os.path.exists(video_id):
+            self._download_webpage(dl_url, video_id)
         self._create_tmp_dir(video_id)
 
         # Extract basic metadata
         metadata_url = video_website + '/presentation/' + video_id + '/metadata.xml'
-        metadata = self._download_xml(metadata_url, video_id)
+        metadata_local_path = video_id + '/metadata.xml'
+        metadata = self._download_and_backup_xml(metadata_url, metadata_local_path)
 
         shapes_url = video_website + '/presentation/' + video_id + '/shapes.svg'
-        shapes = self._download_xml(shapes_url, video_id)
+        shapes_local_path = video_id + '/shapes.svg'
+        shapes = self._download_and_backup_xml(shapes_url, shapes_local_path)
 
         cursor_url = video_website + '/presentation/' + video_id + '/cursor.xml'
-        cursor_infos = self._download_xml(cursor_url, video_id)
+        cursor_local_path = video_id + '/cursor.xml'
+        cursor_infos = self._download_and_backup_xml(cursor_url, cursor_local_path)
 
         # Parse metadata.xml
         meta = metadata.find('./meta')
@@ -297,11 +322,17 @@ class BBBDL(InfoExtractor):
             )
 
         self.to_screen("Downloading slides")
-        self._write_slides(slides_infos, self.ydl)
-        self._write_slides(bonus_images, self.ydl)
-        if add_annotations:
+        self._write_slides(slides_infos)
+        self._write_slides(bonus_images)
+
+        if backup:
+            self.to_screen("Backup Finished")
+            self.to_screen("You can run bbb-dl again to generate the video based on the backed up files!")
+            return
+
+        if self.add_annotations:
             slides_infos = self._add_annotations(slides_infos)
-        if add_cursor:
+        if self.add_cursor:
             slides_infos = self._add_cursor(slides_infos, cursor_infos)
 
         for slides_info in slides_infos:
@@ -326,12 +357,12 @@ class BBBDL(InfoExtractor):
             self.report_warning("Final Slideshow already exists. Abort!")
             return
 
-        if add_webcam:
+        if self.add_webcam:
             self.ffmpeg.mux_slideshow_with_webcam(slideshow_path, webcams_path, webcam_w, webcam_h, result_path)
         else:
             self.ffmpeg.mux_slideshow(slideshow_path, webcams_path, result_path)
 
-        if not keep_tmp_files:
+        if not self.keep_tmp_files:
             self.to_screen("Cleanup")
             self._remove_tmp_dir(video_id)
 
@@ -356,27 +387,65 @@ class BBBDL(InfoExtractor):
         except (OSError, IOError) as err:
             self.ydl.report_error('unable to remove directory ' + error_to_compat_str(err))
 
-    def _write_slides(self, slides_infos: {}, ydl: YoutubeDL):
+    def _download_and_backup_xml(self, url: str, local_path: str):
+        filePath = encodeFilename(local_path)
+        if os.path.exists(filePath):
+            self.to_screen('XML file %s is already present' % (local_path))
+        else:
+            self.to_screen('Downloading XML file %s...' % (url))
+            try_num = 1
+            while True:
+                try:
+                    url_f = self.ydl.urlopen(url)
+                    with open(filePath, 'wb') as xml_f:
+                        shutil.copyfileobj(url_f, xml_f)
+                    self.to_screen('Successfully downloaded to: %s' % (local_path))
+                    break
+                except (compat_urllib_error.URLError, compat_http_client.HTTPException, socket.error) as err:
+                    if os.path.exists(filePath):
+                        os.remove(filePath)
+                    self.report_warning(
+                        f'(Try {try_num} of {self.global_retries}) Unable to download XML file "{url}":'
+                        + f' {error_to_compat_str(err)}'
+                    )
+                    if try_num == self.global_retries:
+                        self.report_warning('XML files are essential. Abort! Please try again later!')
+                        exit(1)
+                    try_num += 1
+        try:
+            tree_root = ElementTree.parse(filePath).getroot()
+        except ParseError as err:
+            self.report_warning('Unable to parse XML file "%s": %s' % (url, error_to_compat_str(err)))
+            self.report_warning('XML files are essential. Abort! Please try again later!')
+            exit(1)
+        return tree_root
+
+    def _write_slides(self, slides_infos: {}):
         for slide in slides_infos:
-            if os.path.exists(encodeFilename(slide.path)):
+            filePath = encodeFilename(slide.path)
+            if os.path.exists(filePath):
                 self.to_screen('Slide %s is already present' % (slide.filename))
             else:
                 self.to_screen('Downloading slide %s...' % (slide.filename))
                 try_num = 1
-                try:
-                    url_f = ydl.urlopen(slide.url)
-                    with open(encodeFilename(slide.path), 'wb') as slide_f:
-                        shutil.copyfileobj(url_f, slide_f)
-                    self.to_screen('Successfully downloaded to: %s' % (slide.path))
-                except (compat_urllib_error.URLError, compat_http_client.HTTPException, socket.error) as err:
-                    self.report_warning(
-                        '(Try %s of 3) Unable to download slide "%s": %s'
-                        % (try_num, slide.url, error_to_compat_str(err))
-                    )
-                    if try_num == 3:
-                        self.report_warning('Slides are essential. Abort! Please try again later!')
-                        exit(1)
-                    try_num += 1
+                while True:
+                    try:
+                        url_f = self.ydl.urlopen(slide.url)
+                        with open(filePath, 'wb') as slide_f:
+                            shutil.copyfileobj(url_f, slide_f)
+                        self.to_screen('Successfully downloaded to: %s' % (slide.path))
+                        break
+                    except (compat_urllib_error.URLError, compat_http_client.HTTPException, socket.error) as err:
+                        if os.path.exists(filePath):
+                            os.remove(filePath)
+                        self.report_warning(
+                            f'(Try {try_num} of {self.global_retries}) Unable to download XML file "{slide.url}":'
+                            + f' {error_to_compat_str(err)}'
+                        )
+                        if try_num == self.global_retries:
+                            self.report_warning('Slides are essential. Abort! Please try again later!')
+                            exit(1)
+                        try_num += 1
 
     def _add_annotations(self, slides_infos: []):
         """Expandes the slides_infos with all annotation slides"""
@@ -436,8 +505,7 @@ class BBBDL(InfoExtractor):
                         )
                     )
 
-                    self.strip_namespace(svg_root)
-                    self.convert_svg_to_png(ElementTree.tostring(svg_root), slide.width, slide.height, new_path)
+                    self.convert_svg_to_png(svg_root, slide.width, slide.height, new_path)
 
                     annotation_slides.append(
                         Slide(
@@ -538,7 +606,7 @@ class BBBDL(InfoExtractor):
 
                     self._paint_cursor(slide, l_x_percent, l_y_percent, new_path)
 
-                    # self.convert_svg_to_png(ElementTree.tostring(svg_root), slide.width, slide.height, new_path)
+                    # self.convert_svg_to_png(svg_root, slide.width, slide.height, new_path)
 
                 tmp_ts_out = ts_out
                 if slide.ts_out < ts_out:
@@ -736,17 +804,55 @@ class BBBDL(InfoExtractor):
         self.ffmpeg.concat_videos(video_list, slideshow_path)
         return slideshow_path
 
-    def convert_svg_to_png(self, svg_bytes, width, height, output_path):
+    def convert_svg_to_png(self, svg_root, width: int, height: int, output_path: str):
         if os.path.isfile(output_path):
             return
-        body = f"""
-        <html style="width: {width}px;height: {height}px;">
-          <body>{svg_bytes.decode()}</body>
-        </html>
-        """
-        Html2Image(output_path=os.path.dirname(output_path)).screenshot(
-            html_str=body, save_as=os.path.basename(output_path), size=(width, height)
-        )
+
+        use_html2image = False
+        if svg_root.find(_s("./svg:foreignObject")) is not None:
+            use_html2image = True
+        svg_bytes = ElementTree.tostring(svg_root)
+
+        if use_html2image:
+            self.strip_namespace(svg_root)
+            body = f"""
+            <html style="width: {width}px;height: {height}px;">
+            <body>{svg_bytes.decode()}</body>
+            </html>
+            """
+
+            Html2Image(
+                output_path=os.path.dirname(output_path),
+                browser_executable=self.chrome_executable,
+                verbose=self.verbose,
+                verbose_chrome=self.verbose_chrome,
+            ).screenshot(html_str=body, save_as=os.path.basename(output_path), size=(width, height))
+        else:
+            PNGSurface.convert(
+                bytestring=svg_bytes,
+                width=width,
+                height=height,
+                write_to=open(output_path, 'wb'),
+            )
+
+
+class Timer(object):
+    '''
+    Timing Context Manager
+    Can be used for future speed comparisons, like this:
+
+    with Timer() as t:
+        Do.stuff()
+    print(f'Do.stuff() took:\t {t.duration:.3f} \tseconds.')
+    '''
+
+    def __enter__(self):
+        self.start = time.perf_counter_ns()
+        return self
+
+    def __exit__(self, *args):
+        end = time.perf_counter_ns()
+        self.duration = (end - self.start) * 10**-9  # 1 nano-sec = 10^-9 sec
 
 
 def get_parser():
@@ -773,20 +879,28 @@ def get_parser():
         help='add the annotations of the professor to the final video',
     )
 
-    """
     parser.add_argument(
         '-ac',
         '--add-cursor',
         action='store_true',
-        help='add the cursor of the professor to the final video',
+        help='add the cursor of the professor to the final video [Experimental, very slow, untested]',
     )
-    """
 
+    parser.add_argument(
+        '-bk',
+        '--backup',
+        action='store_true',
+        help=(
+            'downloads all the content from the server and then stops. After using this option, you can run bbb-dl again to create the video based on the saved files'
+        ),
+    )
     parser.add_argument(
         '-kt',
         '--keep-tmp-files',
         action='store_true',
-        help=('keep the temporary files after finish'),
+        help=(
+            'keep the temporary files after finish. In case of an error bbb-dl will reuse the already generated files'
+        ),
     )
 
     parser.add_argument(
@@ -795,6 +909,21 @@ def get_parser():
         action='store_true',
         help=('print more verbose debug informations'),
     )
+
+    parser.add_argument(
+        '-vc',
+        '--verbose-chrome',
+        action='store_true',
+        help=('print more verbose debug informations of the chrome browser that is used to generate screenshots'),
+    )
+
+    parser.add_argument(
+        '--chrome-executable',
+        type=str,
+        default=None,
+        help='Optional path to your installed Chrome executable (Use it if the path is not detected automatically)',
+    )
+
     parser.add_argument(
         '-ncc',
         '--no-check-certificate',
@@ -836,11 +965,19 @@ def main(args=None):
     parser = get_parser()
     args = parser.parse_args(args)
 
-    BBBDL(args.verbose, args.no_check_certificate, args.encoder, args.audiocodec).run(
-        args.URL,
+    BBBDL(
+        args.verbose,
+        args.no_check_certificate,
+        args.encoder,
+        args.audiocodec,
         args.add_webcam,
         args.add_annotations,
-        False,
+        args.add_cursor,
         args.keep_tmp_files,
+        args.verbose_chrome,
+        args.chrome_executable,
+    ).run(
+        args.URL,
         args.filename,
+        args.backup,
     )
